# -*- coding: utf-8 -*-
# Copyright (c) Sebastian Klaassen. All Rights Reserved.
# Distributed under the MIT License. See LICENSE file for more info.

import abc
import functools
import traceback
from PyQt5.QtWidgets import QApplication
from PyQt5.QtCore import QTimer, QThread, QObject, Qt
from PyQt5.Qt import pyqtSlot, Q_ARG, QCoreApplication
import asyncframes


class EventLoopMeta(type(QObject), abc.ABCMeta):
	pass
class EventLoop(asyncframes.AbstractEventLoop, QObject, metaclass=EventLoopMeta):
    """An implementation of AbstractEventLoop based on PyQt5.

    Args:
        gui_enabled (bool, optional): Defaults to True. If `False`, Qt will run without UI. This argument applies
                                      to the whole application and will be ignored for subsequent event loops.
    """

    qt = None

    def __init__(self, gui_enabled=True):
        # Start Qt
        if EventLoop.qt is None:
            EventLoop.qt = QCoreApplication.instance() or QApplication([]) if gui_enabled else QCoreApplication([])

        asyncframes.AbstractEventLoop.__init__(self)
        QObject.__init__(self)
        self.moveToThread(QThread.currentThread())

    def _run(self):
        try:
            try:
                # Execute QThread event loop
                self.loop = QThread.currentThread()
                self.loop.exec_() # If the current thread is the main thread, this will throw a RuntimeError
            except RuntimeError:
                # Execute QApplication event loop
                self.loop = EventLoop.qt
                EventLoop.qt.exec_()
        except: # pragma: no cover
            print(traceback.format_exc()) # pragma: no cover

    def _stop(self):
        self.loop.exit()

    def _close(self):
        pass

    def _clear(self):
        EventLoop.qt.removePostedEvents(None)

    def _post(self, delay, callback, args):
        QTimer.singleShot(1000 * delay, functools.partial(callback, *args))
    
    def _invoke(self, delay, callback, args):
        self.metaObject().invokeMethod(self, "_invoke_slot", Qt.QueuedConnection, Q_ARG(float, delay), Q_ARG(object, functools.partial(callback, *args)))

    @pyqtSlot(float, object)
    def _invoke_slot(self, delay, callback):
        QTimer.singleShot(1000 * delay, callback)

    def _spawnthread(self, target, args):
        class Thread(QThread):
            def __init__(self, target, args):
                super().__init__()
                self.target = target
                self.args = args
            def run(self):
                try:
                    self.target(*self.args)
                except:
                    print(traceback.format_exc()) # pragma: no cover
        thread = Thread(target, args)
        thread.start()
        return thread

    def _jointhread(self, thread):
        thread.wait()



# -*- coding: utf-8 -*-
# Copyright (c) Sebastian Klaassen. All Rights Reserved.
# Distributed under the MIT License. See LICENSE file for more info.

import abc
import collections
import collections.abc
import datetime
import enum
import inspect
import logging
import sys
import threading
import os
import queue
import warnings


__all__ = [
    'all_', 'animate', 'any_', 'Awaitable', 'AbstractEventLoop', 'Event',
    'find_parent', 'Frame', 'FrameStartupBehaviour', 'FreeEventArgs',
    'get_current_eventloop_index', 'InvalidOperationException', 'hold',
    'PFrame', 'Primitive', 'sleep'
]
__version__ = '2.0.0'


class ThreadLocals(threading.local):
    def __init__(self):
        self.__dict__['_current_eventloop'] = None
        self.__dict__['_current_frame'] = None
_THREAD_LOCALS = ThreadLocals()

class FrameStartupBehaviour(enum.Enum):
    delayed = 1
    immediate = 2

class InvalidOperationException(Exception):
    """Raised when operations are performed out of context.

    Args:
        msg (str): Human readable string describing the exception.
    """

    def __init__(self, msg):
        super().__init__(msg)

class FreeEventArgs(object):
    def __init__(self):
        self.cancel = False

class _AtomicCounter(object):
    """A thread-safe counter that calls a function whenever it hits zero.

        Args:
            initial_value (int): The initial counter value.
            on_zero (Callable): The function to call whenever the counter hits zero.
        """

    def __init__(self, initial_value, on_zero):
        self._lock = threading.Lock()
        self._value = initial_value
        self.on_zero = on_zero
        self.on_zero_args = ()

    def add(self, f):
        """Increment the counter by the given amount."""

        with self._lock:
            self._value += f
            if self._value != 0: return
        self.on_zero(*self.on_zero_args)

    def sub(self, f):
        """Decrement the counter by the given amount."""

        with self._lock:
            self._value -= f
            if self._value != 0: return
        self.on_zero(*self.on_zero_args)

class AbstractEventLoop(metaclass=abc.ABCMeta):
    """Abstract base class of event loops."""

    @abc.abstractmethod
    def _run(self):
        """Start the newly created or stopped event loop.

        This method must be overwritten by the concrete eventloop class.

        Events can be posted or invoked before the eventloop is started.
        """

        raise NotImplementedError # pragma: no cover

    @abc.abstractmethod
    def _stop(self):
        """Stop the running event loop.

        This method must be overwritten by the concrete eventloop class.
        """

        raise NotImplementedError # pragma: no cover

    @abc.abstractmethod
    def _close(self):
        """Close the stopped event loop.

        This method must be overwritten by the concrete eventloop class.

        The eventloop will not be restarted and no events will be posted or invoked after this method is called.
        """

        raise NotImplementedError # pragma: no cover

    @abc.abstractmethod
    def _clear(self):
        """Clear all pending events from the stopped event loop.

        This method must be overwritten by the concrete eventloop class.
        """

        raise NotImplementedError # pragma: no cover

    @abc.abstractmethod
    def _post(self, delay, callback, args):
        """Execute the given callback after ``delay`` seconds.

        This method must be overwritten by the concrete eventloop class.

        This function is **not** threadsafe. It is only called from the thread that this eventloop was started on.
        See :meth:`AbstractEventLoop._invoke` for the threadsafe version of this method.

        Args:
            delay (float): The time to wait before executing the callback.
            callback (function): The function to be called.
            args (tuple): The arguments to pass to the callback.
        """

        raise NotImplementedError # pragma: no cover

    @abc.abstractmethod
    def _invoke(self, delay, callback, args):
        """Execute the given callback after ``delay`` seconds.

        This method must be overwritten by the concrete eventloop class.

        This function **is** threadsafe. It can be called from any thread.
        See :meth:`AbstractEventLoop._post` for the unsafe version of this method.

        Args:
            delay (float): The time to wait before executing the callback.
            callback (function): The function to be called.
            args (tuple): The arguments to pass to the callback.
        """

        raise NotImplementedError # pragma: no cover

    def _spawnthread(self, target, args):
        """Create and start a daemonic worker thread.

        By default, this will create and start a ``threading.Thread``.
        The concrete eventloop class can overwrite this method to use a different threading model.

        Args:
            target (function): Same as the ``target`` argument of ``threading.Thread``.
            args (tuple): Same as the ``args`` argument of ``threading.Thread``.

        Returns:
            The created thread.
        """

        thread = threading.Thread(target=target, args=args, daemon=True)
        thread.start()
        return thread

    def _jointhread(self, thread):
        """Wait until the given thread terminates.

        Only threads created with :meth:`AbstractEventLoop._spawnthread` are awaited with this method.

        Args:
            thread: The thread to wait for.
        """

        thread.join()

    def __init__(self):
        self._idle = True
        self._eventloop_affinity = self
        self._result = None
        self._exception = None

    def run(self, frame, *frameargs, num_threads=0, **framekwargs):
        if num_threads <= 0:
            num_threads = len(os.sched_getaffinity(0))
        if _THREAD_LOCALS._current_eventloop is not None:
            raise InvalidOperationException("Another event loop is already running")
        _THREAD_LOCALS._current_eventloop = self

        eventloop_queue = queue.Queue()
        self.event_queue = queue.Queue()
        def worker_thread(parent_eventloop):
            eventloop = parent_eventloop.__class__()
            eventloop.event_queue = parent_eventloop.event_queue
            _THREAD_LOCALS._current_eventloop = eventloop
            eventloop_queue.put(eventloop)
            eventloop._run()
            eventloop._close()
        workers = [self._spawnthread(target=worker_thread, args=(self,)) for i in range(num_threads - 1)]

        # Collect an array of all eventloops and distribute that array among all eventloops
        self.eventloops = [self]
        for worker in workers: self.eventloops.append(eventloop_queue.get())
        for eventloop in self.eventloops[1:]: eventloop.eventloops = self.eventloops

        self._idle = False
        self._result = None
        self._exception = None

        try:
            mainframe = frame(*frameargs, **framekwargs)
            if not mainframe.removed:
                mainframe._listeners.add(self) # Listen to mainframe finished event
                self._run()
        except:
            raise
        else:
            if self._exception:
                raise self._exception
            else:
                return self._result
        finally:
            _THREAD_LOCALS._current_eventloop = None
            _THREAD_LOCALS._current_frame = None

            # Clear event queue
            while True:
                try:
                    self.event_queue.get(False)
                except queue.Empty:
                    break

            # Clear main eventloop
            self._clear()
            self._idle = True

            # Stop worker eventloops
            for eventloop in self.eventloops[1:]: eventloop._invoke(0, eventloop._stop, ())
            for worker in workers: self._jointhread(worker)

    def _enqueue(self, delay, callback, args, eventloop_affinity=None):
        if len(self.eventloops) == 1: # If running singlethreaded, ...
            # Execute callback from current eventloop
            if _THREAD_LOCALS._current_eventloop == self:
                self._post(delay, callback, args)
            else:
                self._invoke(delay, callback, args)
        elif eventloop_affinity: # If a target eventloop was provided, ...
            # Execute callback from target eventloop
            if _THREAD_LOCALS._current_eventloop == eventloop_affinity:
                eventloop_affinity._post(delay, callback, args)
            else:
                eventloop_affinity._invoke(delay, callback, args)
        else: # If no target eventloop was provided, ...
            if delay > 0.0:
                # Call _enqueue again with 0 delay after 'delay' seconds
                #TODO: Consider running a dedicated event loop instead of eventloops[-1] for delays
                if _THREAD_LOCALS._current_eventloop == self.eventloops[-1]:
                    self.eventloops[-1]._post(delay, self.eventloops[-1]._enqueue, (0.0, callback, args))
                else:
                    self.eventloops[-1]._invoke(delay, self.eventloops[-1]._enqueue, (0.0, callback, args))
            else: # If delay == 0, ...
                # Place the callback on the event queue
                self.event_queue.put((callback, args))

                # Wake up an idle event (if any)
                for eventloop in self.eventloops:
                    if eventloop._idle:
                        eventloop._idle = False
                        eventloop._invoke(0, eventloop._dequeue, ())
                        break

    def _dequeue(self):
        try:
            callback, args = self.event_queue.get_nowait()
        except queue.Empty:
            self._idle = True
        else:
            callback(*args)
            if not self.event_queue.empty():
                self._post(0, self._dequeue, ())
            else:
                self._idle = True

    @staticmethod
    def sendevent(eventsource, event, process_counter=None, blocking=False):
        # Save current frame, since it will be modified inside Awaitable.process()
        currentframe = _THREAD_LOCALS._current_frame

        try:
            eventsource.process(eventsource, event, process_counter, blocking)
        finally:
            # Restore current frame
            _THREAD_LOCALS._current_frame = currentframe

    def postevent(self, eventsource, event, delay=0):
        self._enqueue(delay, AbstractEventLoop.sendevent, (eventsource, event, None, False), eventsource._eventloop_affinity)

    def process(self, sender, msg, process_counter=None, blocking=False):
        self._result = msg
        self._stop() # Stop event loop


class Awaitable(collections.abc.Awaitable):
    """An awaitable frame or event.

    Every node in the frame hierarchy is a subclass of `Awaitable`. An awaitable has a `__name__`,
    a parent awaitable (None, if the awaitable is the main frame), a list of child awaitables and
    a result, that gets set when the awaitable finishes.

    Args:
        name (str): The name of the awaitable.
    """

    def __init__(self, name):
        self.__name__ = name
        self._parent = None
        self._removed = False
        self._result = None
        self._listeners = set()
        self._eventloop_affinity = None
        self.exception_handler = None

    def _remove(self, process_counter=None, blocking=False, ondone=lambda result: None):
        # Mark self as removed
        if self._removed:
            ondone(False)
            if process_counter:
                process_counter.sub(1)
            return
        self._removed = True

        # Remove self from parent frame
        if self._parent:
            self._parent._children.remove(self)

        # Wake up listeners
        if self._listeners:
            listeners = self._listeners
            self._listeners = set()

            if blocking:
                if process_counter:
                    process_counter.add(len(listeners))

                for listener in listeners:
                    if listener._eventloop_affinity is None or listener._eventloop_affinity == _THREAD_LOCALS._current_eventloop:
                        listener.process(self, self._result, process_counter, blocking)
                    else:
                        listener._eventloop_affinity._invoke(0, listener.process, (self, self._result, process_counter, blocking))
            else:
                for listener in listeners:
                    _THREAD_LOCALS._current_eventloop._enqueue(0, listener.process, (self, self._result), listener._eventloop_affinity)

        del self
        ondone(True)
        if process_counter:
            process_counter.sub(1)
        return

    def remove(self):
        """Remove this awaitable from the frame hierarchy.

        Returns:
            bool: If `True`, this event was removed. If `False` the request was either canceled, or the event had already been removed before
        """

        remove_event = Event(str(self.__name__) + ".remove", singleshot=True)
        process_counter = _AtomicCounter(1, lambda result: AbstractEventLoop.sendevent(remove_event, result, None, True)) # Add process: Frame.remove
        process_counter.on_zero_args = (True,) # Default remove() result to True
        def onremoved(result):
            process_counter.on_zero_args = (result,)
        self._remove(process_counter, False, onremoved)
        return remove_event

    @property
    def removed(self):
        """Boolean property, indicating whether this awaitable has been removed from the frame hierarchy."""
        return self._removed

    def __str__(self):
        """Human readable representation of this frame."""
        return self.__name__

    def __repr__(self):
        return "<{}.{} object at 0x{:x}>".format(self.__module__, self.__name__, id(self))

    def __lt__(self, other):
        """Make awaitables sortable by name."""
        return self.__name__ < other.__name__

    def __await__(self):
        if self.removed: # If this awaitable already finished
            return self._result
        try:
            return (yield self)
        except (StopIteration, GeneratorExit):
            return self._result

    @abc.abstractmethod
    def _step(self, sender, msg):
        raise NotImplementedError # pragma: no cover

    def process(self, sender, msg, process_counter=None, blocking=False):
        """Propagate an event from its source and along awaiting nodes through the frame hierarchy.

        Args:
            sender (Awaitable): The source of the event or an awaited node that woke up.
            msg: The incomming event or propagated message.
        """

        _THREAD_LOCALS._current_frame = self # Activate self
        try:
            self._step(sender, msg)
        except BaseException as err:
            if getattr(self, 'ready', True): # If self is ready or self doesn't have a ready event
                # Send ready event to all listeners that have a ready event, but aren't ready yet
                for listener in self._listeners:
                    if not getattr(listener, 'ready', True):
                        AbstractEventLoop.sendevent(listener.ready, None, None, True)

            # Store result
            if type(err) == StopIteration:
                self._result = err.value
            elif type(err) != GeneratorExit:
                self._result = err

                # Call exception handler
                if err != msg:
                    awaitable = self
                    while awaitable:
                        if awaitable.exception_handler:
                            try:
                                awaitable.exception_handler(self, err)
                            except Exception as exception_handler_err:
                                err = exception_handler_err
                            else:
                                break
                        awaitable = awaitable._parent
                    if not awaitable:
                        maineventloop = _THREAD_LOCALS._current_eventloop.eventloops[0]
                        maineventloop._exception = err
                        if maineventloop._eventloop_affinity is None or maineventloop._eventloop_affinity == _THREAD_LOCALS._current_eventloop:
                            maineventloop._stop()
                        else:
                            maineventloop._eventloop_affinity._invoke(0, maineventloop._stop, ())

            # Remove awaitable and propagate event
            self._remove(process_counter, blocking)
        else:
            if getattr(self, 'ready', True): # If self is ready or self doesn't have a ready event
                # Send ready event to all listeners that have a ready event, but aren't ready yet
                for listener in self._listeners:
                    if not getattr(listener, 'ready', True):
                        AbstractEventLoop.sendevent(listener.ready, None, None, True)

            if process_counter:
                process_counter.sub(1)

    def __and__(self, other):
        """Register A & B as shortcut for ``all_(A, B)``.

        Args:
            other (Awaitable): The other awaitable.

        Returns:
            all\_: A frame awaiting `self` and `other`.
        """

        return all_(self, other)
    def __or__(self, other):
        """Register A | B as shortcut for ``any_(A, B)``.

        Args:
            other (Awaitable): The other awaitable.

        Returns:
            any\_: A frame awaiting `self` or `other`.
        """

        return any_(self, other)

class Event(Awaitable):
    """An awaitable event.

    Instantiate or overload this class to implement new events.
    Each type of event should be emitted by exactly one event class.
    For example, key-up and key-down events should be implemented by two separate events.
    Events represent leave nodes in the frame hierarchy.

    Args:
        name (str): The name of the event.
        singleshot (bool, optional): Defaults to False. If `True`, removes the event after it has been woken.
    """

    def __init__(self, name, singleshot=False):
        super().__init__(name)
        self.singleshot = singleshot
        self.eventloop = _THREAD_LOCALS._current_eventloop # Store creating eventloop, as a fallback in case self.post() is called from a thread without an eventloop

    def _remove(self, process_counter=None, blocking=False, ondone=lambda result: None):
        if self.singleshot:
            super()._remove(process_counter, blocking, ondone)
        else:
            # Wake up listeners
            if self._listeners:
                listeners = self._listeners
                self._listeners = set()

                if blocking:
                    if process_counter is not None:
                        process_counter.add(len(listeners))

                    for listener in listeners:
                        if listener._eventloop_affinity is None or listener._eventloop_affinity == _THREAD_LOCALS._current_eventloop:
                            listener.process(self, self._result, process_counter, blocking)
                        else:
                            listener._eventloop_affinity._invoke(0, listener.process, (self, self._result, process_counter, blocking))
                else:
                    for listener in listeners:
                        _THREAD_LOCALS._current_eventloop._enqueue(0, listener.process, (self, self._result), listener._eventloop_affinity)
            ondone(False)
            if process_counter:
                process_counter.sub(1)

    def __bool__(self):
        return self._removed

    def _step(self, sender, msg):
        """Handle incoming events.

        Args:
            sender (Event): The event to be handled. This value is always identical to `self`.
            msg: The event arguments.

        Raises:
            StopIteration: If this event should wake up awaiting frames, raise a StopIteration with `value` set to the event arguments.
        """

        stop = StopIteration()
        stop.value = msg
        raise stop

    def send(self, args=None):
        """Dispatch and immediately process an event.

        Args:
            args (optional): Defaults to None. Event arguments, for example, the progress value on a progress-update event.
        """

        send_event = Event(str(self.__name__) + ".send", singleshot=True)
        process_counter = _AtomicCounter(1, lambda: AbstractEventLoop.sendevent(send_event, None, None, True))
        AbstractEventLoop.sendevent(self, args, process_counter, True)
        return send_event

    def post(self, args=None, delay=0):
        """Enqueue an event in the event loop.

        Args:
            args (optional): Defaults to None. Event arguments, for example, the progress value on a progress-update event.
            delay (float, optional): Defaults to 0. The time in seconds to wait before posting the event.
        """

        (_THREAD_LOCALS._current_eventloop or self.eventloop).postevent(self, args, delay)

class all_(Awaitable):
    """An awaitable that blocks the awaiting frame until all passed awaitables have woken up.

    Args:
        awaitables (Awaitable[]): A list of all awaitables to await.
    """

    def __init__(self, *awaitables):
        super().__init__("all({})".format(", ".join(str(a) for a in awaitables)))
        self._remove_lock = threading.Lock()

        self._awaitables = collections.defaultdict(list)
        self._result = [None] * len(awaitables)
        for i, awaitable in enumerate(awaitables):
            if awaitable.removed:
                self._result[i] = awaitable._result
            else:
                self._awaitables[awaitable].append(i)
                awaitable._listeners.add(self)

        if not self._awaitables:
            self._remove()
            return

        self._parent = _THREAD_LOCALS._current_frame
        if self._parent:
            self._parent._children.append(self)

    def _step(self, sender, msg):
        """Respond to an awaking child.

        Args:
            sender (Awaitable): The awaking child.
            msg: The awaking child's result or an exception raised in a child frame.

        Raises:
            StopIteration: Once all children woke up, this raises a StopIteration with `value` set to a dict of all children's results.
            BaseException: If msg is an Exception other than GeneratorExit or StopIteration, the exception is re-raised.
        """

        for i in self._awaitables.pop(sender, ()):
            self._result[i] = msg

        if not self._awaitables:
            stop = StopIteration()
            stop.value = self._result
            raise stop

    def _remove(self, process_counter=None, blocking=False, ondone=lambda result: None):
        if self.removed:
            ondone(False)
            if process_counter:
                process_counter.sub(1)
            return

        with self._remove_lock:
            if self.removed: # If this frame was closed while acquiring the lock, ...
                ondone(False)
                if process_counter:
                    process_counter.sub(1)
                return

            for awaitable in self._awaitables:
                awaitable._listeners.discard(self)
            self._awaitables.clear()

            # Remove awaitable
            super()._remove(process_counter, blocking, ondone)
            return
        if process_counter:
            process_counter.sub(1)
        ondone(False)

class any_(Awaitable):
    """An awaitable that blocks the awaiting frame until any of the passed awaitables wakes up.

    Args:
        awaitables (Awaitable[]): A list of all awaitables to await.
    """

    def __init__(self, *awaitables):
        super().__init__("any({})".format(", ".join(str(a) for a in awaitables)))
        self._remove_lock = threading.Lock()

        self._awaitables = set()
        for awaitable in awaitables:
            if awaitable.removed:
                self._result = (awaitable, awaitable._result)
                self._remove()
                return
            else:
                self._awaitables.add(awaitable)
                awaitable._listeners.add(self)

        self._parent = _THREAD_LOCALS._current_frame
        if self._parent:
            self._parent._children.append(self)

    def _step(self, sender, msg):
        """Respond to an awaking child.

        Args:
            sender (Awaitable): The awaking child.
            msg: The awaking child's result or an exception raised in a child frame.

        Raises:
            StopIteration: If msg indicates an awaking child, store its result as this frame's result.
            BaseException: Forward any exceptions.
        """

        stop = StopIteration()
        stop.value = (sender, msg)
        raise stop

    def _remove(self, process_counter=None, blocking=False, ondone=lambda result: None):
        if self.removed:
            ondone(False)
            if process_counter:
                process_counter.sub(1)
            return

        with self._remove_lock:
            if self.removed: # If this frame was closed while acquiring the lock, ...
                ondone(False)
                if process_counter:
                    process_counter.sub(1)
                return

            for awaitable in self._awaitables:
                awaitable._listeners.discard(self)
            self._awaitables.clear()

            # Remove awaitable
            super()._remove(process_counter, blocking, ondone)
            return
        if process_counter:
            process_counter.sub(1)
        ondone(False)


class sleep(Event):
    """An awaitable event used for suspending execution by the specified amount of time.

    A duration of 0 seconds will resume the awaiting frame as soon as possible.
    This is useful to implement non-blocking loops.

    Args:
        seconds (float, optional): Defaults to 0. The duration to wait.
    """

    def __init__(self, seconds=0.0):
        super().__init__("sleep({})".format(seconds), singleshot=True)

        # Raise event
        self.post(None, max(0, seconds))

class hold(Event):
    """An awaitable event used for suspending execution indefinitely.

    Frames are automatically removed when the frame coroutine finishes.
    If you would like the frame to remain open until it is removed, write `await hold()` at the end of the coroutine.
    """

    def __init__(self):
        super().__init__("hold()", singleshot=True)
    def _step(self, sender, msg):
        """ Ignore any incoming events."""

        pass

class animate(Event):
    """An awaitable event used for periodically calling a callback function for the specified amount of time.

    Args:
        seconds (float): The duration of the animation.
        callback (Callable[float, None]): The function to be called on every iteration. The first parameter of `callback` indicates animation progress between 0 and 1.
        interval (float, optional): Defaults to 0.0. The minimum time in seconds between two consecutive calls of the callback.
    """

    def __init__(self, seconds, callback, interval=0.0):
        super().__init__("animate()", singleshot=True)
        self.seconds = seconds
        self.callback = callback
        self.interval = interval
        self.startTime = datetime.datetime.now()
        self._final_event = False

        # Raise event
        if seconds <= interval:
            self._final_event = True
            self.post(None, max(0, seconds))
        else:
            self.post(None, interval)

    def _step(self, sender, msg):
        """Resend the animation event until the timeout is reached."""
        t = (datetime.datetime.now() - self.startTime).total_seconds()

        if t >= self.seconds or self._final_event:
            self.callback(1.0)
            stop = StopIteration()
            stop.value = msg
            raise stop
        else:
            self.callback(t / self.seconds)
            t = (datetime.datetime.now() - self.startTime).total_seconds() # Recompute t after callback

            # Reraise event
            if self.seconds - t <= self.interval:
                self._final_event = True
                self.post(None, max(0, self.seconds - t))
            else:
                self.post(None, self.interval)

class FrameMeta(abc.ABCMeta):
    def __new__(mcs, name, bases, dct):
        frameclass = super().__new__(mcs, name, bases, dct)
        if bases != (Awaitable,): # If frameclass != Frame, ...
            frameclass.Factory = type(frameclass.__name__ + '.Factory', (frameclass.Factory,), {}) # Derive factory from base class factory
        frameclass.Factory.frameclass = frameclass
        return frameclass

class Frame(Awaitable, metaclass=FrameMeta):
    """An object within the frame hierarchy.

    This class represents the default frame class. All other frame classes have
    to be derived from :class:`Frame`.

    A frame is an instance of a frame class. Use the nested Factory class to
    create frames.

    The factory class is created by decorating a function or coroutine with
    ``@FRAME``, where ``FRAME`` is the frame class.

    Example: ::

        class MyFrameClass(Frame):
            pass

        @MyFrameClass
        async def my_frame_factory():
            pass

        assert(type(my_frame_factory) == MyFrameClass.Factory)
        my_frame = my_frame_factory()
        assert(type(my_frame) == MyFrameClass)

    Args:
        startup_behaviour (FrameStartupBehaviour, optional): Defaults to FrameStartupBehaviour.delayed.
            Controls whether the frame is started immediately or queued on the eventloop.
        thread_idx (int, optional): Defaults to None. If set, forces the scheduler to affiliate this frame with the given thread.

    Raises:
        ValueError: If `thread_idx` is outside the range of allocated threads.

            The number of allocated threads is controlled by the `num_threads` parameter of :meth:`AbstractEventLoop.run`.
    """

    class Factory(object):
        """A frame function declared in the context of a frame class.
        
        Args:
            framefunc (Callable): The function or coroutine that describes the frame's behaviour.
            frameclassargs (tuple): Positional arguments to the frame class.
            frameclasskwargs (dict): Keyword arguments to the frame class.
        """

        def __init__(self, framefunc, frameclassargs, frameclasskwargs):
            self.framefunc = framefunc
            self.frameclassargs = frameclassargs
            self.frameclasskwargs = frameclasskwargs

        def __call__(self, *frameargs, **framekwargs):
            """Produce an instance of the frame.
            
            Raises:
                InvalidOperationException: Raised when no event loop is currently running.
            
            Returns:
                Frame: The newly created frame instance.
            """

            if _THREAD_LOCALS._current_eventloop is None:
                raise InvalidOperationException("Can't call frame without a running event loop")
            frame = super(Frame, self.__class__.frameclass).__new__(self.__class__.frameclass)
            frame.__init__(*self.frameclassargs, **self.frameclasskwargs)
            frame.create(self.framefunc, *frameargs, **framekwargs)
            return frame

    def __new__(cls, *frameclassargs, **frameclasskwargs):
        if len(frameclassargs) == 1 and not frameclasskwargs and callable(frameclassargs[0]): # If @frame was called without parameters
            framefunc = frameclassargs[0]
            frameclassargs = ()
            return cls.Factory(framefunc, frameclassargs, frameclasskwargs)
        else: # If @frame was called with parameters
            return lambda framefunc: cls.Factory(framefunc, frameclassargs, frameclasskwargs)

    def __init__(self, startup_behaviour=FrameStartupBehaviour.delayed, thread_idx=None):
        if thread_idx is not None and (thread_idx < 0 or thread_idx >= len(_THREAD_LOCALS._current_eventloop.eventloops)):
            raise ValueError("thread_idx must be an index between 0 and " + str(len(_THREAD_LOCALS._current_eventloop.eventloops)))
        super().__init__(self.__class__.__name__)
        self.startup_behaviour = startup_behaviour
        self._parent = _THREAD_LOCALS._current_frame
        if self._parent:
            self._parent._children.append(self)
        self._children = []
        self._activechild = None
        self._primitives = []
        self._generator = None
        self._generator_eventloop = None
        self._freeing = False
        self.ready = Event(str(self.__name__) + ".ready", True)
        self.ready.ready = self.ready # Set ready state of ready event to itself. This way the ready event will propagate through `await frame.ready`
        self.free = Event(str(self.__name__) + ".free", False)
        self._eventloop_affinity = _THREAD_LOCALS._current_eventloop
        if thread_idx is not None:
            self._eventloop_affinity = self._eventloop_affinity.eventloops[thread_idx]
        self._remove_lock = threading.Lock()

    def create(self, framefunc, *frameargs, **framekwargs):
        """Start the frame function with the given arguments.

        Args:
            framefunc (function): A coroutine or regular function controlling the behaviour of this frame.
                                If `framefunc` is a coroutine, then the frame only exists until the coroutine exits.
        """

        if framefunc and not self.removed and self._generator is None:
            self.__name__ = framefunc.__name__
            self.free.__name__ = str(self.__name__) + ".free"

            # Activate self
            _THREAD_LOCALS._current_frame = self

            hasself = 'self' in inspect.signature(framefunc).parameters
            self._generator = framefunc(self, *frameargs, **framekwargs) if hasself else framefunc(*frameargs, **framekwargs)

            if inspect.isawaitable(self._generator): # If framefunc is a coroutine
                if self.startup_behaviour == FrameStartupBehaviour.delayed:
                    # Post coroutine to the event queue
                    _THREAD_LOCALS._current_eventloop.postevent(self, None, 0.0)
                elif self.startup_behaviour == FrameStartupBehaviour.immediate:
                    # Start coroutine
                    try:
                        self._step(None, None)
                    except (StopIteration, GeneratorExit):
                        pass
                    finally:
                        # Activate parent
                        _THREAD_LOCALS._current_frame = self._parent
                else:
                    raise ValueError('startup_behaviour must be FrameStartupBehaviour.delayed or FrameStartupBehaviour.immediate')

            # Activate parent
            _THREAD_LOCALS._current_frame = self._parent

    def _step(self, sender, msg):
        """Resume the frame coroutine.

        Args:
            sender (Awaitable): The resumed awaitable.
            msg: A message to be forwarded to the coroutine or None to start the coroutine.

        Raises:
            StopIteration: Raised if the coroutine finished with a result.
            GeneratorExit: Raised if the coroutine finished without a result.
            Exception: Raised if the coroutine encountered an error.
        """

        if self.removed:
            stop = StopIteration()
            stop.value = self._result
            raise stop

        if self._generator is not None:
            self._generator_eventloop = _THREAD_LOCALS._current_eventloop
            try:
                awaitable = self._generator.send(msg) # Continue coroutine
            except (StopIteration, GeneratorExit): # If frame finished
                self._generator_eventloop = None
                if msg is None: AbstractEventLoop.sendevent(self.ready, None, None, True) # Send ready event if frame finished without ever being awaited
                raise # Propagate event
            except Exception as err: # If frame raised exception
                self._generator_eventloop = None
                raise # Propagate event
            else: # If frame awaits awaitable
                self._generator_eventloop = None
                awaitable._listeners.add(self) # Listen to events of awaitable

                # Send ready event if not yet ready and awaitable is ready or doesn't have a ready event
                if getattr(awaitable, 'ready', True) and not self.ready: # If awaitable is ready or awaitable doesn't have a ready event
                    AbstractEventLoop.sendevent(self.ready, None, None, True)

    def _mark_freeing(self, value):
        self._freeing = value
        for child in self._children:
            if isinstance(child, Frame):
                child._mark_freeing(value)

    def _send_free_event(self, free_args, process_counter):
        process_counter.add(len(self._children))

        # Send child frame free events
        for child in self._children:
            if isinstance(child, Frame):
                child._send_free_event(free_args, process_counter)
            else:
                process_counter.sub(1)

        # Send current frame free event
        AbstractEventLoop.sendevent(self.free, free_args, process_counter, True)

    def _remove(self, process_counter=None, blocking=False, ondone=lambda result: None):
        if self._freeing: # If this frame's free event is being processed, ...
            self._remove_stage2(process_counter, blocking, ondone) # Remove the frame
        else: # If this frame's free event wasn't send yet, ...
            self._mark_freeing(True) # Mark this frame and all children for removal

            free_args = FreeEventArgs()
            def after_free():
                if free_args.cancel: # If the free event was canceled, ...
                    self._mark_freeing(False) # Unmark this frame and all children from removal
                    ondone(False) # Awake awaiting frames with value 'False'
                    if process_counter:
                        process_counter.sub(1)
                elif self._removed: # If this frame was removed as a result of the free event, ...
                    ondone(True) # Awake awaiting frames with value 'True'
                    if process_counter:
                        process_counter.sub(1)
                else: # If the free event wasn't canceled and didn't remove this frame, ...
                    self._remove_stage2(process_counter, blocking, ondone) # Remove this frame
            self._send_free_event(free_args, _AtomicCounter(1, after_free))

    def _remove_stage2(self, process_counter=None, blocking=False, ondone=lambda result: None):
        if self.removed or not self._remove_lock.acquire(blocking=False): # If this frame was closed in response to the free event or removal is already in progress, ...
            ondone(False)
            if process_counter:
                process_counter.sub(1) # Remove process: Frame._remove_stage2
        else:
            try:
                # Remove child frames
                genexit = None
                if process_counter:
                    process_counter.add(len(self._children))
                while self._children:
                    try:
                        self._children[-1]._remove(process_counter, blocking)
                    except GeneratorExit as err:
                        genexit = err # Delay raising GeneratorExit

                # Remove primitives
                while self._primitives:
                    self._primitives[-1].remove()

                # Stop framefunc
                if self._generator: # If framefunc is a coroutine
                    genevtlp = self._generator_eventloop # Avoid race condition between if and invoke
                    if genevtlp == _THREAD_LOCALS._current_eventloop: # If the coroutine is running on the current eventloop
                        # Calling coroutine.close() from within the coroutine is illegal, so we throw a GeneratorExit manually instead
                        try:
                            self._generator.close()
                        except ValueError:
                            genexit = GeneratorExit()
                        self._generator = None
                    elif genevtlp is not None: # If the coroutine is running on another eventloop
                        if process_counter:
                            process_counter.add(1) # Add process: Frame._remove_stage3
                        genevtlp._invoke(0, self._remove_stage3, (process_counter,))
                    else: # If the coroutine isn't running
                        self._generator.close()
                        self._generator = None

                # Remove awaitable
                super()._remove(process_counter, blocking, ondone)
            finally:
                self._remove_lock.release()

                # Raise delayed GeneratorExit exception
                if genexit:
                    raise genexit

    def _remove_stage3(self, process_counter=None):
        genevtlp = self._generator_eventloop # Avoid race condition between if and invoke
        if genevtlp is None or genevtlp == _THREAD_LOCALS._current_eventloop: # If the coroutine isn't running or it's running on the current eventloop
            self._generator.close()
            if process_counter:
                process_counter.sub(1) # Remove process: Frame._remove_stage3
        else: # If the coroutine is running on another eventloop
            genevtlp._invoke(0, self._remove_stage3, (process_counter,))

class PFrame(Frame):
    """A parallel :class:`Frame` that can run on any thread.

    Multithreading can be enabled for any frame by changing its base class to
    :class:`PFrame`.

    The only difference between :class:`Frame` and :class:`PFrame` is that
    instances of :class:`PFrame` are not restricted to run on the same thread as
    their parent frame.

    Args:
        startup_behaviour (FrameStartupBehaviour, optional): Defaults to FrameStartupBehaviour.delayed.
            Controls whether the frame is started immediately or queued on the eventloop.
        thread_idx (int, optional): Defaults to None. If set, forces the scheduler to affiliate this frame with the given thread.

    Raises:
        ValueError: If `thread_idx` is outside the range of allocated threads.

            The number of allocated threads is controlled by the `num_threads` parameter of :meth:`AbstractEventLoop.run`.
    """

    def __init__(self, startup_behaviour=FrameStartupBehaviour.delayed, thread_idx=None):
        super().__init__(startup_behaviour, thread_idx)
        if thread_idx is None:
            self._eventloop_affinity = None


class Primitive(object):
    """An object owned by a frame of the specified frame class.

    A primitive has to be created within the frame function of its owner or within the frame function of any child frame of its owning frame class.
    If it is created within a child frame, it will still be registered with the closest parent of the owning frame class.

    Args:
        owner (class): The owning frame class.

    Raises:
        TypeError: Raised if owner is not a frame class.
        Exception: Raised if a primitive is created outside the frame function of its owning frame class.
    """

    def __init__(self, owner):
        self._removed = False

        # Validate parameters
        if not issubclass(owner, Frame):
            raise TypeError("'owner' must be of type Frame")

        # Find parent frame of class 'owner'
        self._owner = find_parent(owner)
        if not self._owner:
            raise InvalidOperationException(self.__class__.__name__ + " can't be defined outside " + owner.__name__)

        # Register with parent frame
        self._owner._primitives.append(self)

    def remove(self):
        """Remove this primitive from its owner.

        Returns:
            bool: If `True`, this event was removed. If `False` the request was either canceled, or the event had already been removed before
        """

        if self._removed:
            return False
        self._removed = True
        self._owner._primitives.remove(self)
        return True


def get_current_eventloop_index():
    return _THREAD_LOCALS._current_eventloop.eventloops.index(_THREAD_LOCALS._current_eventloop) if getattr(_THREAD_LOCALS, '_current_eventloop', None) and hasattr(_THREAD_LOCALS._current_eventloop, 'eventloops') else None

def find_parent(parenttype):
    parent = _THREAD_LOCALS._current_frame
    while parent and not issubclass(type(parent), parenttype):
        parent = parent._parent
    return parent



# -*- coding: utf-8 -*-
# Copyright (c) Sebastian Klaassen. All Rights Reserved.
# Distributed under the MIT License. See LICENSE file for more info.

import os.path
import setuptools
import asyncframes

def read(filename, begin_after=None):
    with open(os.path.join(os.path.dirname(__file__), filename)) as file:
        content = file.read()

    if begin_after and begin_after in content:
        content = content[content.find(begin_after) + len(begin_after):]

    return content

setuptools.setup(
    name='asyncframes',
    packages=['asyncframes'],
    version=asyncframes.__version__,
    description='Parallel programming for software engineers',
    long_description=read('README.rst', "----\n\n") + '\n\n' + read('CHANGES.rst'),
    author='Sebastian Klaassen',
    author_email='rcsepp@hotmail.com',
    license='MIT',
    url='https://github.com/RcSepp/asyncframes',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries :: Application Frameworks'
    ],
    python_requires='>=3.5'
)



# -*- coding: utf-8 -*-
# Copyright (c) Sebastian Klaassen. All Rights Reserved.
# Distributed under the MIT License. See LICENSE file for more info.

import asyncio
import asyncframes


class EventLoop(asyncframes.AbstractEventLoop):
    """An implementation of AbstractEventLoop based on asyncio."""

    def __init__(self):
        super().__init__()
        try:
            self.loop = asyncio.get_event_loop() # Use existing eventloop
        except RuntimeError: # If no eventloop exists on this thread, ...
            self.loop = asyncio.new_event_loop() # Create a new eventloop
            asyncio.set_event_loop(self.loop) # Make the new eventloop current

    def _run(self):
        self.loop.run_forever()

    def _stop(self):
        self.loop.stop()

    def _close(self):
        self.loop.close()

    def _clear(self):
        # Re-open asyncio eventloop to discard any pending events
        self.loop.close() # Close current event loop
        self.loop = asyncio.new_event_loop() # Create a new eventloop
        asyncio.set_event_loop(self.loop) # Make the new eventloop current

    def _post(self, delay, callback, args):
        if not self.loop.is_closed():
            if delay <= 0:
                self.loop.call_soon(callback, *args)
            else:
                self.loop.call_later(delay, callback, *args)

    def _invoke(self, delay, callback, args):
        if not self.loop.is_closed():
            if delay <= 0:
                self.loop.call_soon_threadsafe(callback, *args)
            else:
                self.loop.call_soon_threadsafe(self.loop.call_later, delay, callback, *args)



# -*- coding: utf-8 -*-
# Copyright (c) Sebastian Klaassen. All Rights Reserved.
# Distributed under the MIT License. See LICENSE file for more info.

import datetime
import io
import logging
import math
import threading
import queue
import re
import time
import unittest
import asyncframes
from asyncframes import *
from asyncframes import _THREAD_LOCALS
import line_tracer

# Uncomment the following line to run all tests using PFrames
# Frame = PFrame

# NUM_ITERATIONS controls the number of times each test is executed
NUM_ITERATIONS = 1

# NUM_THREADS controls the number of worker threads to start when running an event loop
# If 0, the number of CPU threads is used (according to os.sched_getaffinity(0))
# If 1, no multithreading will be performed
NUM_THREADS = 4

# If true, any exceptions during an iteration of a test case will print a full line-by-line trace of the executed test
USE_TRACER = False
TRACE_OUTPUT = "./trace.txt" # Output file for trace output or None to print to standard output
TRACE_MODE = line_tracer.Trace.Mode.on_error


class nullcontext(object):
    """A context manager that does nothing.

    See https://docs.python.org/3/library/contextlib.html#contextlib.nullcontext
    Implemented here for Python versions before 3.7
    """

    def __init__(self, enter_result=None):
        self.enter_result = enter_result
    def __enter__(self):
        return self.enter_result
    def __exit__(self, exc_type, exc_value, traceback):
        return

class MyFrame(Frame):
    @staticmethod
    def mystaticmethod(test):
        test.log.debug('static method called')
    classvar = 'class variable'

@MyFrame
async def wait(test, seconds, name):
    result = await sleep(seconds)
    test.log.debug(name)
    return "some result"

class MyException(Exception):
    pass

EVENTLOOP_CLASS = None
SKIP_TEST_CASE = None

class TestAsyncFrames(unittest.TestCase):
    def setUp(self):
        # Create default event loop if no event loop was created by a base class
        if not hasattr(self, 'loop'):
            from asyncframes.asyncio_eventloop import EventLoop
            self.loop = EventLoop()

        # Announce event loop if different
        global EVENTLOOP_CLASS
        if self.loop.__class__ != EVENTLOOP_CLASS:
            print()
            print("Using {}.{}".format(self.loop.__class__.__module__, self.loop.__class__.__name__))
            EVENTLOOP_CLASS = self.loop.__class__

        # Create logger for debugging program flow using time stamped log messages
        # Create time stamped log messages using self.log.debug(...)
        # Test program flow by passing an expected_log to self.run_frame()
        self.log = logging.getLogger(self._testMethodName + str(threading.get_ident()))
        self.log.setLevel(logging.DEBUG)
        self.logstream = io.StringIO()
        loghandler = logging.StreamHandler(self.logstream)
        loghandler.setLevel(logging.DEBUG)
        class TimedFormatter(logging.Formatter):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)
                self.starttime = datetime.datetime.now()
            def format(self, record):
                t = (datetime.datetime.now() - self.starttime).total_seconds()
                msg = super().format(record)
                return str(math.floor(t * 10) / 10) + ": " + msg if msg else str(math.floor(t * 10) / 10)
        self.logformatter = TimedFormatter("%(message)s")
        loghandler.setFormatter(self.logformatter)
        self.log.addHandler(loghandler)

    def run_frame(self, frame, *frameargs, expected_log=None, assert_raises=None, assert_raises_regex=None):
        @Frame
        async def mainframe():
            self.logstream.truncate(0) # Reset log
            self.logstream.seek(0) # Reset log
            self.logformatter.starttime = datetime.datetime.now() # Reset log start time
            return await frame(*frameargs)
        result = None
        cm = line_tracer.Trace(TRACE_MODE, TRACE_OUTPUT, True) if USE_TRACER else nullcontext()
        for _ in range(NUM_ITERATIONS):
            with cm:
                try:
                    result = self.loop.run(mainframe, num_threads=NUM_THREADS)
                    self.log.debug('done')
                    if expected_log is not None:
                        # Compare log with expected_log
                        expected = '\n'.join(line.strip() for line in expected_log.strip('\n').split('\n')) # Remove leading and trailing empty lines and white space
                        self.assertEqual(expected, self.logstream.getvalue())
                except Exception as err:
                    if assert_raises:
                        failed = type(err) != assert_raises
                    elif assert_raises_regex:
                        failed = type(err) != assert_raises_regex[0] or not re.match(assert_raises_regex[1], str(err))
                    else:
                        failed = True
                    if failed:
                        raise
                else:
                    if assert_raises or assert_raises_regex:
                        raise AssertionError((assert_raises or assert_raises_regex[0]).__name__ + " not raised")
        return result

    def test_simple(self):
        test = self
        @Frame
        async def main():
            await wait(test, 0.1, '1')
        test.run_frame(main, expected_log="""
            0.1: 1
            0.1: done
        """)

    def test_regular_function_mainframe(self):
        test = self
        @Frame
        async def remove_after(frame, seconds):
            await sleep(seconds)
            frame.remove()
        @Frame
        def main(self):
            remove_after(self, 0.1)
        test.run_frame(main, expected_log="""
            0.1: done
        """)

    def test_negative_sleep_duration(self):
        test = self
        @MyFrame
        async def main():
            await sleep(-1)
        test.run_frame(main, expected_log="""
            0.0: done
        """)

    def test_await_order_1(self):
        test = self
        @MyFrame
        async def main():
            wait(test, 0.1, '1')
            await wait(test, 0.2, '2')
        test.run_frame(main, expected_log="""
            0.1: 1
            0.2: 2
            0.2: done
        """)

    def test_await_order_2(self):
        test = self
        @MyFrame
        async def main():
            await wait(test, 0.1, '1')
            await wait(test, 0.2, '2')
        test.run_frame(main, expected_log="""
            0.1: 1
            0.3: 2
            0.3: done
        """)

    def test_await_order_3(self):
        test = self
        @MyFrame
        async def main():
            await wait(test, 0.1, '1')
            wait(test, 0.2, '2')
        test.run_frame(main, expected_log="""
            0.1: 1
            0.1: done
        """)

    def test_await_order_4(self):
        test = self
        @MyFrame
        async def main():
            wait(test, 0.1, '1')
            wait(test, 0.2, '2')
        test.run_frame(main, expected_log="""
            0.0: done
        """)

    def test_await_order_5(self):
        test = self
        @MyFrame
        async def main():
            w1 = wait(test, 0.1, '1')
            w2 = wait(test, 0.2, '2')
            await (w1 & w2)
        test.run_frame(main, expected_log="""
            0.1: 1
            0.2: 2
            0.2: done
        """)

    def test_await_order_6(self):
        test = self
        @MyFrame
        async def main():
            w1 = wait(test, 0.1, '1')
            w2 = wait(test, 0.2, '2')
            await (w1 | w2)
        test.run_frame(main, expected_log="""
            0.1: 1
            0.1: done
        """)

    def test_frame_result(self):
        test = self
        @MyFrame
        async def main():
            test.log.debug(await wait(test, 0.1, '1'))
        test.run_frame(main, expected_log="""
            0.1: 1
            0.1: some result
            0.1: done
        """)
    
    def test_blocking_sleep(self):
        test = self
        @MyFrame
        async def main():
            test.log.debug(await sleep(0.1))
        test.run_frame(main, expected_log="""
            0.1: None
            0.1: done
        """)
    
    def test_non_blocking_sleep(self):
        test = self
        @MyFrame
        async def main():
            test.log.debug(await sleep(0))
        test.run_frame(main, expected_log="""
            0.0: None
            0.0: done
        """)

    def test_staticmethod(self):
        test = self
        @MyFrame
        async def main():
            MyFrame.mystaticmethod(test)
        test.run_frame(main, expected_log="""
            0.0: static method called
            0.0: done
        """)

    def test_classvar(self):
        test = self
        @MyFrame
        async def main():
            test.log.debug(MyFrame.classvar)
        test.run_frame(main, expected_log="""
            0.0: class variable
            0.0: done
        """)

    def test_primitive(self):
        test = self
        class MyPrimitive(Primitive):
            def __init__(self):
                super().__init__(MyFrame)
        class MyPrimitive2(Primitive):
            def __init__(self):
                super().__init__(MyPrimitive)
        class MyFrame2(Frame):
            pass

        @MyFrame
        async def f1():
            MyPrimitive()
        test.run_frame(f1)

        @MyFrame2
        async def f2():
            MyPrimitive()
        test.run_frame(f2, assert_raises=InvalidOperationException)

        @MyFrame
        async def f3():
            await f2()
        test.run_frame(f3)

        @MyFrame
        async def f4():
            with test.assertRaises(TypeError):
                MyPrimitive2()
        test.run_frame(f4)

        with test.assertRaises(InvalidOperationException):
            MyPrimitive()

    def test_frameclassargs(self):
        test = self
        class MyFrame2(Frame):
            def __init__(self, param, kwparam):
                super().__init__()
                test.log.debug("param=%s" % param)
                test.log.debug("kwparam=%s" % kwparam)
        @MyFrame2('param_value', kwparam='kwparam_value')
        async def main():
            pass
        test.run_frame(main, expected_log="""
            0.0: param=param_value
            0.0: kwparam=kwparam_value
            0.0: done
        """)

    def test_Frame_current(self):
        test = self
        @MyFrame
        async def frame(self):
            test.subframe_counter = 0
            test.assertEqual(_THREAD_LOCALS._current_frame, self)
            async_subframe() # Test passive async frame
            test.assertEqual(_THREAD_LOCALS._current_frame, self)
            await async_subframe() # Test active async frame
            test.assertEqual(_THREAD_LOCALS._current_frame, self)
            subframe() # Test passive frame
            test.assertEqual(_THREAD_LOCALS._current_frame, self)
            @Frame
            async def remove_after(frame, seconds):
                await sleep(seconds)
                frame.remove()
            sf = subframe()
            remove_after(sf, 0.1)
            await sf # Test active frame
            test.assertEqual(_THREAD_LOCALS._current_frame, self)
        @MyFrame
        async def async_subframe(self):
            test.assertEqual(_THREAD_LOCALS._current_frame, self)
            await sleep()
            test.subframe_counter += 1
        @MyFrame
        def subframe(self):
            test.assertEqual(_THREAD_LOCALS._current_frame, self)
            test.subframe_counter += 1
        test.run_frame(frame)
        test.assertEqual(test.subframe_counter, 4)

    def test_remove_self(self):
        test = self
        @Frame
        def frame(self):
            test.log.debug("1")
            self.remove()
            test.log.debug("2") # Frame.remove() doesn't interrupt regular frame functions
        test.run_frame(frame, expected_log="""
            0.0: 1
            0.0: 2
            0.0: done
        """)
        @Frame
        async def async_frame(self):
            test.log.debug("1")
            self.remove()
            test.log.debug("never reached") # Frame.remove() interrupts async frame functions
        test.run_frame(async_frame, expected_log="""
            0.0: 1
            0.0: done
        """)

    def test_pframe_remove(self):
        """Test removing self after handling a free event on another thread.

        Expected behaviour:
        After waking all frames that wait for frame_a.free, frame_a.remove
        should terminate frame_a's coroutine by raising GeneratorExit.

        Faced issue:
        After handling frame_a.free inside frame_b, frame_a raises GeneratorExit
        from thread 3. GeneratorExit should instead be raised from the thread
        that frame_a's coroutine is running on (thread 2).
        """

        test = self
        @Frame(thread_idx=2)
        async def frame_a(self):
            time.sleep(0.1)
            r = self.remove()
            time.sleep(0.1)
            await r
            test.log.debug("never reached") # Frame.remove() interrupts async frame functions
        @Frame(thread_idx=3)
        async def frame_b(a):
            await a.free
        @Frame(thread_idx=1)
        async def main(self):
            a = frame_a()
            await (a & frame_b(a))
        test.run_frame(main, expected_log="""
            0.2: done
        """)

    def test_reremove(self):
        test = self
        class MyPrimitive(Primitive):
            def __init__(self):
                super().__init__(Frame)
        @Frame
        async def remove_after(frame, seconds):
            await sleep(seconds)
            test.log.debug('Removing frame')
            test.assertEqual(await frame.remove(), True)
            test.log.debug('Re-removing frame')
            test.assertEqual(await frame.remove(), False)
        @Frame
        async def main(self):
            self.p = None
            @Frame
            async def primitive_owner():
                self.p = MyPrimitive()
                await sleep(0.0)
                test.log.debug('Removing primitive')
            primitive_owner()

            remove_after(self, 0.3)
            a = (self.free | sleep(0.1))
            await a
            test.log.debug('Re-removing any_')
            test.assertEqual(await a.remove(), False)

            s = sleep(0.1)
            a = (self.free & s)
            await a
            test.log.debug('Re-removing all_')
            test.assertEqual(await a.remove(), False)

            test.log.debug('Re-removing event source')
            test.assertEqual(await s.remove(), False)

            test.log.debug('Re-removing primitive')
            test.assertEqual(self.p.remove(), False)

            test.log.debug('Re-removing frame during frame.free event')
            test.assertEqual(await self.remove(), True)
        test.run_frame(main, expected_log="""
            0.0: Removing primitive
            0.1: Re-removing any_
            0.3: Removing frame
            0.3: Re-removing all_
            0.3: Re-removing event source
            0.3: Re-removing primitive
            0.3: Re-removing frame during frame.free event
            0.3: Re-removing frame
            0.3: done
        """)

    def test_awaited_by_multiple(self):
        test = self
        @Frame
        async def waitfor(w):
            await w
        @Frame
        async def main1(self):
            w = wait(test, 0.1, '1')
            await (w & w)
        test.run_frame(main1)
        @Frame
        async def main2(self):
            w = wait(test, 0.1, '2')
            await (w | w)
        test.run_frame(main2)
        @Frame
        async def main3(self):
            w = sleep(0.1)
            a1 = waitfor(w)
            a2 = waitfor(w)
            await (a1 & a2)
        test.run_frame(main3)
        @Frame
        async def main4(self):
            w = wait(test, 0.1, '3')
            a1 = waitfor(w)
            a2 = waitfor(w)
            await (a1 & a2)
        test.run_frame(main4)
        @Frame
        async def main5(self):
            w = wait(test, 0.1, '4')
            a1 = waitfor(w)
            a2 = waitfor(w)
            await (a1 | a2)
        test.run_frame(main5)
        @Frame
        async def main6(self):
            s = sleep(0.1)
            await (s | s & s)
        test.run_frame(main6)

    def test_finished_before_await(self):
        test = self
        @Frame
        async def main():
            s = sleep(0.1)
            w = wait(test, 0.1, '1')
            await sleep(0.2)
            await s
            test.log.debug(await w)
            awaitable_repr = Awaitable.__repr__
            Awaitable.__repr__ = Awaitable.__str__
            test.log.debug(await (s | w))
            test.log.debug(await (s & w))
            Awaitable.__repr__ = awaitable_repr
        test.run_frame(main, expected_log="""
            0.1: 1
            0.2: some result
            0.2: (sleep(0.1), None)
            0.2: [None, 'some result']
            0.2: done
        """)

    def test_custom_event(self):
        test = self
        @Frame
        async def main(self):
            ae = Event('my event')

            send_event(0.1, ae)
            send_event(0.2, ae)
            sender, args = await ae
            test.log.debug("'%s' raised '%s' with args '%s'", sender, ae, args)
            sender, args = await ae
            test.log.debug("'%s' reraised '%s' with args '%s'", sender, ae, args)

            post_event(0.1, ae)
            ae.post((self, 'my event args'), 0.2)
            sender, args = await ae
            test.log.debug("'%s' raised '%s' with args '%s'", sender, ae, args)
            sender, args = await ae
            test.log.debug("'%s' reraised '%s' with args '%s'", sender, ae, args)

            threading.Thread(target=invoke_event, args=(0.1, ae)).start()
            threading.Thread(target=ae.post, args=((self, 'my event args'), 0.2)).start()
            sender, args = await ae
            test.log.debug("'%s' raised '%s' with args '%s'", sender, ae, args)
            sender, args = await ae
            test.log.debug("'%s' reraised '%s' with args '%s'", sender, ae, args)
        @Frame
        async def send_event(self, seconds, awaitable_event):
            await sleep(seconds)
            awaitable_event.send((self, 'my event args'))
        @Frame
        async def post_event(self, seconds, awaitable_event):
            await sleep(seconds)
            awaitable_event.post((self, 'my event args'))
        def invoke_event(seconds, awaitable_event):
            time.sleep(seconds)
            awaitable_event.post(('invoke_event', 'my event args'))

        test.run_frame(main, expected_log="""
            0.1: 'send_event' raised 'my event' with args 'my event args'
            0.2: 'send_event' reraised 'my event' with args 'my event args'
            0.3: 'post_event' raised 'my event' with args 'my event args'
            0.4: 'main' reraised 'my event' with args 'my event args'
            0.5: 'invoke_event' raised 'my event' with args 'my event args'
            0.6: 'main' reraised 'my event' with args 'my event args'
            0.6: done
        """)

    def test_exceptions(self):
        test = self
        test.maxDiff = None

        @Frame
        async def raise_immediately():
            raise MyException()
        @Frame
        async def raise_delayed():
            await sleep(0.1)
            raise MyException()
        @Frame
        async def return_exception():
            return MyException()
        def suppressing_exception_handler(frame, err):
            if type(err) == MyException:
                test.log.debug("Suppressed {} of frame {}".format(repr(err), frame))
            else:
                raise
        def reraising_exception_handler(frame, err):
            test.log.debug("Reraised {} of frame {}".format(repr(err), frame))
            raise

        # Test eventloop raising exception
        test.run_frame(raise_immediately, assert_raises=MyException)
        # Test eventloop returning exception
        test.assertEqual(type(test.run_frame(return_exception)), MyException)

        # Test event handling within eventloop
        @Frame
        async def main(self):
            self.exception_handler = suppressing_exception_handler
            for frame in [raise_immediately, raise_delayed, return_exception]:
                frame()
                hold() | frame()
                sleep() & frame()
                test.assertEqual(type(await frame()), MyException)
                test.assertEqual(tuple(map(type, await (hold() | frame()))), (frame.frameclass, MyException))
                test.assertEqual(tuple(map(type, await (sleep() & frame()))), (type(None), MyException))

            self.exception_handler = reraising_exception_handler
            frame = raise_delayed()
            frame.exception_handler = suppressing_exception_handler
            await frame

            #TODO: # Implement starting frames with exception handler
            #TODO:
            #TODO: * Option 1:
            #TODO:   with exception_handler(suppressing_exception_handler):
            #TODO:       raise_immediately()
            #TODO:
            #TODO: * Option 2:
            #TODO:   raise_immediately(exception_handler=suppressing_exception_handler)
            raise_immediately() #TODO: ... instead of this line
            await sleep(0.2)
        test.run_frame(main, assert_raises=MyException, expected_log="""
            0.0: Suppressed MyException() of frame raise_immediately
            0.0: Suppressed MyException() of frame raise_immediately
            0.0: Suppressed MyException() of frame raise_immediately
            0.0: Suppressed MyException() of frame raise_immediately
            0.0: Suppressed MyException() of frame raise_immediately
            0.0: Suppressed MyException() of frame raise_immediately
            0.0: Suppressed MyException() of frame raise_delayed
            0.0: Suppressed MyException() of frame raise_delayed
            0.0: Suppressed MyException() of frame raise_delayed
            0.0: Suppressed MyException() of frame raise_delayed
            0.0: Suppressed MyException() of frame raise_delayed
            0.0: Suppressed MyException() of frame raise_delayed
            0.0: Suppressed MyException() of frame raise_delayed
            0.0: Reraised MyException() of frame raise_immediately
        """)

    def test_animate(self):
        test = self
        @Frame
        async def a():
            await animate(0.25, lambda f: test.log.debug(''), 0.1)
        @Frame
        async def main():
            await a()
        test.run_frame(main, expected_log="""
            0.1
            0.2
            0.2
            0.2: done
        """)

    def test_unfinished_await(self):
        test = self
        @MyFrame
        async def frame():
            await subframe()
            await (subframe() | hold())
            await (subframe() & sleep())
        @MyFrame
        async def subframe():
            await sleep()
            await sleep()
        test.run_frame(frame)

    def test_rerun(self):
        test = self
        @Frame
        async def main():
            await wait(test, 0.1, 'main')
        @Frame
        def raise_exception():
            raise MyException
        test.run_frame(main, expected_log="""
            0.1: main
            0.1: done
        """)
        test.run_frame(main, expected_log="""
            0.1: main
            0.1: done
        """)
        test.run_frame(raise_exception, assert_raises=MyException)
        test.run_frame(main, expected_log="""
            0.1: main
            0.1: done
        """)

    def test_meta(self):
        test = self
        @Frame
        async def main(self):
            test.assertEqual(str(self), "main")
            test.assertRegex(repr(self), r"<asyncframes.main object at 0x\w*>")
        test.run_frame(main)

    def test_invalid_usage(self):
        test = self
        @Frame
        def raise_already_running():
            test.loop.run(wait, test, 0.0, '', num_threads=NUM_THREADS)
        test.run_frame(raise_already_running, assert_raises_regex=(InvalidOperationException, "Another event loop is already running"))
        with test.assertRaisesRegex(InvalidOperationException, "Can't call frame without a running event loop"):
            wait(test, 0.0, '')

    def test_reawait(self):
        test = self
        @Frame
        async def main():
            s1 = sleep(0.1)
            s2 = sleep(0.2)
            s3 = sleep(0.3)
            test.assertEqual((await (s1 | s2))[0], s1)
            test.log.debug('1')
            test.assertEqual((await (s2 | s3))[0], s2)
            test.log.debug('2')
            test.assertEqual(await (s2 & s3), [None, None])
            test.log.debug('3')
            await s1
            await s2
            await s3
            test.log.debug('4')
        test.run_frame(main, expected_log="""
            0.1: 1
            0.2: 2
            0.3: 3
            0.3: 4
            0.3: done
        """)

    def test_ready(self):
        test = self
        @Frame
        async def case1():
            test.log.debug('case1_1')
        @Frame
        async def case2():
            test.log.debug('case2_1')
            await hold()
            test.log.debug('case2_2')
        @Frame
        async def case3():
            test.log.debug('case3_1')
            raise MyException()
        @Frame
        async def main(self):
            self.exception_handler = lambda frame, err: test.log.debug("Frame exception caught: %s", repr(err))
            await case1().ready
            await case2().ready
            c3 = case3()
            test.assertEqual(tuple(map(type, await (c3.ready | c3))), (Frame, MyException))
            test.assertFalse(c3.ready)

        test.run_frame(main, expected_log="""
            0.0: case1_1
            0.0: case2_1
            0.0: case3_1
            0.0: Frame exception caught: MyException()
            0.0: done
        """)

    def test_recursive_ready(self):
        test = self
        @Frame
        async def frame1():
            @Frame
            async def frame2():
                @Frame
                async def frame3():
                    @Frame
                    async def frame4():
                        @Frame
                        async def frame5():
                            test.log.debug('frame5 ready')
                            await sleep()
                        test.log.debug('frame4 ready')
                        #TODO: Implement ready propagation through all_
                        # e.g.: await (frame5() & frame5() & sleep())
                        await frame5()
                    test.log.debug('frame3 ready')
                    #TODO: Implement ready propagation through any_
                    # e.g.: await (frame4() | hold())
                    await frame4()
                test.log.debug('frame2 ready')
                await frame3().ready
            test.log.debug('frame1 ready')
            await frame2()
        @Frame
        async def main(self):
            await frame1().ready
        test.run_frame(main, expected_log="""
            0.0: frame1 ready
            0.0: frame2 ready
            0.0: frame3 ready
            0.0: frame4 ready
            0.0: frame5 ready
            0.0: done
        """)

    def test_free(self):
        test = self
        @Frame
        async def frame():
            sf = subframe()
            await sleep(0.1)
            sf.remove()

            sf = subframe()
            await sf.ready
        @Frame
        async def subframe(self):
            test.log.debug('1')
            await self.free
            test.log.debug('2')
        test.run_frame(frame, expected_log="""
            0.0: 1
            0.1: 2
            0.1: 1
            0.1: 2
            0.1: done
        """)

    def test_cancel_free(self):
        test = self
        @Frame
        async def frame():
            sf = subframe()
            await sf.ready
            test.log.debug(await sf.remove()) # This will be canceled by subframe -> False
            test.log.debug(await sf.remove()) # This will also be canceled by subframe -> False
            test.log.debug(await sf.remove()) # This will remove subframe -> True
            test.log.debug(await sf.remove()) # This will have no effect -> False
        @Frame
        async def subframe(self):
            f = await self.free
            f.cancel = True
            f = await self.free
            f.cancel = True
            await self.free
        test.run_frame(frame, expected_log="""
            0.0: False
            0.0: False
            0.0: True
            0.0: False
            0.0: done
        """)

    def test_cancel_child_free(self):
        test = self
        @Frame
        async def frame():
            sf = subframe()
            await sf.ready
            test.log.debug(await sf.remove()) # This will be canceled by subframe_child -> False
            test.log.debug(await sf.remove()) # This will also be canceled by subframe_child -> False
            test.log.debug(await sf.remove()) # This will remove subframe and subframe_child -> True
            test.log.debug(await sf.remove()) # This will have no effect -> False
        @Frame
        async def subframe(self):
            @Frame
            async def subframe_child(self):
                f = await self.free
                f.cancel = True
                f = await self.free
                f.cancel = True
                await self.free
            subframe_child()
            await hold()
        test.run_frame(frame, expected_log="""
            0.0: False
            0.0: False
            0.0: True
            0.0: False
            0.0: done
        """)

    def test_cancel_child_free(self):
        test = self
        @Frame
        async def frame():
            sf = subframe()
            await sf.ready
            test.log.debug(await sf.remove())
            test.log.debug(await sf.remove())
            test.log.debug(await sf.remove())
        @Frame
        async def subframe(self):
            @Frame
            async def subframe_child(self):
                f = await self.free
                f.cancel = True
                await self.free
            await subframe_child().ready
            await hold()
        test.run_frame(frame, expected_log="""
            0.0: False
            0.0: True
            0.0: False
            0.0: done
        """)

    def test_delayed_await(self):
        test = self
        @Frame
        async def main(self):
            self.exception_handler = lambda frame, err: test.log.debug("Frame exception caught: %s", repr(err))
            f = raise_exception()
            await sleep(0.1)
            await f
            test.log.debug('after exception')
        @Frame
        async def raise_exception():
            raise MyException
        test.run_frame(main, expected_log="""
            0.0: Frame exception caught: MyException()
            0.1: after exception
            0.1: done
        """)

        test = self
        @Frame
        async def main():
            f = raise_exception()
            await sleep(0.1)
            await f
            test.log.debug('after exception')
        @Frame
        async def raise_exception():
            raise MyException
        test.run_frame(main, assert_raises=MyException)

    def test_startup_behaviour(self):
        test = self
        @Frame
        async def frame():
            sf = immediate_subframe()
            test.assertEqual(sf.var, 'value')
            await sf.ready
            test.assertEqual(sf.var, 'value')

            sf = delayed_subframe()
            with test.assertRaises(AttributeError):
                test.assertEqual(sf.var, 'value')
            await sf.ready
            test.assertEqual(sf.var, 'value')
        @Frame(startup_behaviour=FrameStartupBehaviour.immediate)
        async def immediate_subframe(self):
            self.var = 'value'
        @Frame(startup_behaviour=FrameStartupBehaviour.delayed)
        async def delayed_subframe(self):
            time.sleep(0.1)
            self.var = 'value'
            await sleep()
        test.run_frame(frame)

    def test_pframe(self):
        test = self
        @PFrame
        async def pframe():
            pframe_threadid = threading.get_ident()
            test.log.debug('pframe_1')
            await sleep(0.1)
            test.log.debug('pframe_2')
            await sleep(0.1)
            test.log.debug('pframe_3')
        @asyncframes.Frame # This has to be a Frame to test successfully. Make sure it's a Frame even when 'Frame' is overwritten with 'PFrame'
        async def main():
            main_threadid = threading.get_ident()
            p = pframe()
            await p.ready
            test.assertEqual(threading.get_ident(), main_threadid)
            test.log.debug('main_1')
            await p
            test.assertEqual(threading.get_ident(), main_threadid)
            test.log.debug('main_2')
        test.run_frame(main, expected_log="""
            0.0: pframe_1
            0.0: main_1
            0.1: pframe_2
            0.2: pframe_3
            0.2: main_2
            0.2: done
        """)

    def test_pframe_send(self):
        test = self
        @PFrame
        async def pframe(e):
            await e
        @PFrame
        async def pframe2(self):
            self.e = Event('my_event')
            await self.e
        @Frame
        async def main():
            e = Event('my_event')
            p = pframe(e)
            await p.ready
            e.send()
            test.assertEqual(p.removed, True)

            p2 = pframe2()
            await p2.ready
            p2.e.send()
            test.assertEqual(p2.removed, True)
        test.run_frame(main)

    def test_await_send(self):
        test = self
        @Frame(thread_idx=2)
        async def a(event):
            await event
            test.log.debug(1)

        @Frame(thread_idx=3)
        async def b(event):
            await a(event)

        @Frame
        async def main():
            event = Event("TODO")
            b(event)
            b(event)
            await sleep(0.1)
            await event.send()
            test.log.debug(2)
        test.run_frame(main, expected_log="""
            0.1: 1
            0.1: 1
            0.1: 2
            0.1: done
        """)

    def test_send_across_threads(self):
        test = self
        @Frame(thread_idx=2)
        async def frame1(e):
            await e
            test.log.debug('1')
        @Frame(thread_idx=3)
        async def frame2(e):
            await e.send()
            test.log.debug('2')
        @Frame
        async def main(self):
            e = Event('my_event')
            f1 = frame1(e)
            await f1.ready
            f2 = frame2(e)
            await (f1 & f2)
        test.run_frame(main, expected_log="""
            0.0: 1
            0.0: 2
            0.0: done
        """)

    def test_ready_across_threads(self):
        test = self
        @Frame(thread_idx=2)
        async def frame1():
            time.sleep(0.1)
            test.log.debug('2')
            await sleep(0.1)
            test.log.debug('4')
        @Frame(thread_idx=3)
        async def frame2(f1):
            test.log.debug('1')
            await f1.ready
            test.log.debug('3')
        @Frame
        async def main(self):
            f1 = frame1()
            f2 = frame2(f1)
            await (f1 & f2)
        test.run_frame(main, expected_log="""
            0.0: 1
            0.1: 2
            0.1: 3
            0.2: 4
            0.2: done
        """)

    def test_free_across_threads(self):
        """Test awaiting the free event from a thread different than the one
        from the removed frame.

        Test status:
        Deterministic

        Expected behaviour:
        When frame f1 is to be removed, all code between `await f1.free` and the
        next await or the end of the frame should be executed before f1 is
        actually removed. This behaviour allows cleanup code to run before the
        frame is removed.

        Faced issue:
        Send is expected to only return once the event has been handled. This
        does not apply when sending across threads. If the awaiting frame runs
        on another thread than the sending frame, send() behaves like post(). It
        would violate the thread-restriction to directly execute the awaiting
        frame. Instead, it posts the request into the awaiting thread's
        eventloop and returns before the posted request is handled.
        """

        test = self
        @Frame(thread_idx=2)
        async def frame1():
            await sleep(0.1)
            test.log.debug('2')
        @Frame(thread_idx=3)
        async def frame2(f1):
            test.log.debug('1')
            await f1.free
            time.sleep(0.1)
            test.log.debug('3 ' + str(f1.removed))
        @Frame
        async def main(self):
            f1 = frame1()
            await frame2(f1)
        test.run_frame(main, expected_log="""
            0.0: 1
            0.1: 2
            0.2: 3 False
            0.2: done
        """)

    def test_remove_across_threads(self):
        """Test removing a frame from another thread."""

        test = self
        @Frame(thread_idx=2)
        async def frame1(self):
            await self.free
            test.log.debug('2 ' + str(self.removed))
        @Frame(thread_idx=3)
        async def frame2(f1):
            test.log.debug(1)
            time.sleep(0.1)
            await f1.remove()
            test.log.debug(3)
        @Frame
        async def main(self):
            f1 = frame1()
            await frame2(f1)
        test.run_frame(main, expected_log="""
            0.0: 1
            0.1: 2 False
            0.1: 3
            0.1: done
        """)

    def test_thread_independence(self):
        test = self
        errors = queue.Queue()
        def test_thread(tc):
            try:
                test.__class__(tc).debug()
            except unittest.SkipTest:
                pass
            except Exception as err:
                errors.put(err)

        testcases = unittest.defaultTestLoader.getTestCaseNames(test.__class__)
        testcases.remove('test_thread_independence')

        threads = [threading.Thread(target=test_thread, args=(testcase,)) for testcase in testcases]
        for thread in threads:
            thread.daemon = True
            thread.start()
            time.sleep(0.1) # Delay test starts to reduce the effect of startup latency on code timings
        endtime = datetime.datetime.now() + datetime.timedelta(seconds=5) # Limit processing time to 5 seconds
        for thread in threads: thread.join(max(0, (endtime - datetime.datetime.now()).total_seconds()))

        for i, thread in enumerate(threads):
            if thread.is_alive():
                print(testcases[i] + " did not finish")

        if not errors.empty():
            raise Exception(str(errors.qsize()) + " test cases failed") from errors.get()

class TestPyQt5EventLoop(TestAsyncFrames):
    def setUp(self):
        # Create PyQt5 event loop
        try:
            from asyncframes.pyqt5_eventloop import EventLoop
        except ImportError:
            # Announce that we skip this test case, if not announced before
            global SKIP_TEST_CASE
            if self.__class__ != SKIP_TEST_CASE:
                print()
                print("Unable to import asyncframes.pyqt5_eventloop. Skipping unit tests for this event loop.")
                SKIP_TEST_CASE = self.__class__

            raise unittest.SkipTest
        else:
            self.loop = EventLoop(gui_enabled=False)
            super().setUp()

    def test_thread_independence(self):
        test = self
        errors = queue.Queue()
        import PyQt5.Qt
        class TestThread(PyQt5.Qt.QThread):
            err = None
            def __init__(self, tc):
                super().__init__()
                self.tc = tc
            def run(self):
                try:
                    test.__class__(self.tc).debug()
                except unittest.SkipTest:
                    pass
                except Exception as err:
                    errors.put(err)

        testcases = unittest.defaultTestLoader.getTestCaseNames(test.__class__)
        testcases.remove('test_thread_independence')

        threads = [TestThread(testcase) for testcase in testcases]
        for thread in threads:
            thread.start()
            time.sleep(0.1) # Delay test starts to reduce the effect of startup latency on code timings
        for thread in threads: thread.wait()

        if not errors.empty():
            raise Exception(str(errors.qsize()) + " test cases failed") from errors.get()

if __name__ == "__main__":
    unittest.main()
